#define com_maroontress_libcgi_util_Tub_IMPORT

#include <stdlib.h>
#include <string.h>
#include <com/maroontress/libcgi/util/Tub.h>

struct Tub {
    char *data;
    size_t size;
    size_t capacity;
    int hasError;
};

struct Tub *
Tub_create(void)
{
    struct Tub *tub;

    if ((tub = (struct Tub *)malloc(sizeof(*tub))) == NULL)
	return NULL;
    tub->data = NULL;
    tub->size = 0;
    tub->capacity = 0;
    tub->hasError = 0;
    return tub;
}

int
Tub_writeCString(struct Tub *tub, const char *str)
{
    return Tub_write(tub, str, strlen(str));
}

static int
expand(struct Tub *tub, size_t nextSize)
{
    static const size_t DELTA = 1024;
    size_t capacity;
    void *data;

    capacity = nextSize + DELTA;
    if ((data = realloc(tub->data, capacity)) == NULL) {
	tub->hasError = 1;
	return -1;
    }
    tub->data = data;
    tub->capacity = capacity;
    return 0;
}

int
Tub_writeByte(struct Tub *tub, unsigned int byte)
{
    unsigned char b = byte;

    if (tub->capacity == tub->size && expand(tub, tub->capacity) < 0)
	return -1;
    tub->data[tub->size] = b;
    tub->size += 1;
    return 0;
}

int
Tub_writeU16(struct Tub *tub, unsigned int u16)
{
    unsigned char c[2];

    c[0] = u16 >> 8;
    c[1] = u16;
    return Tub_write(tub, c, sizeof(c));
}

int
Tub_writeU32(struct Tub *tub, unsigned int u32)
{
    unsigned char c[4];

    c[0] = u32 >> 24;
    c[1] = u32 >> 16;
    c[2] = u32 >> 8;
    c[3] = u32;
    return Tub_write(tub, c, sizeof(c));
}

int
Tub_write(struct Tub *tub, const void *data, size_t size)
{
    size_t nextSize;

    nextSize = tub->size + size;
    if (tub->capacity < nextSize && expand(tub, nextSize) < 0)
	return -1;
    memcpy(tub->data + tub->size, data, size);
    tub->size = nextSize;
    return 0;
}

void
Tub_setSize(struct Tub *tub, size_t size)
{
    if (tub->size <= size)
	return;
    tub->size = size;
}

size_t
Tub_getSize(struct Tub *tub)
{
    return tub->size;
}

void
Tub_free(struct Tub *tub)
{
    if (tub == NULL)
	return;
    free(tub->data);
    free(tub);
}

void
Tub_clear(struct Tub *tub)
{
    tub->size = 0;
    tub->hasError = 0;
}

void *
Tub_createByteArray(struct Tub *tub)
{
    void *bytes;

    if ((bytes = malloc(tub->size)) == NULL)
	return NULL;
    memcpy(bytes, tub->data, tub->size);
    return bytes;
}

int
Tub_hasError(struct Tub *tub)
{
    return tub->hasError;
}
